#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "ntcmd.h"
#include "shmalloc.h"
#ifdef MULTITHREADED
#include "threader.h"
#endif

extern int gSelectOnSysPoll;
unsigned int idCycleTimer = 0;
int cycleTerminated = FALSE;
int cycleStarted = FALSE;
int cycle_busy = 0;
int gSystemDelay = 20;
char gTineCycleWindowName[256];

HANDLE hTineCycleWnd = 0;
#define hNULL 0
void DoEvents(int maxevents);
int hasMessages(void);

void srvTmrTask(void)
{
  if (cycleTerminated) return;
  if (cycle_busy++) return;
  SystemCycle(FALSE);
  cycle_busy = 0;
}

TINE_EXPORT int openIPCPipe(char *pipeName)
{
  char ipcfn[32];
  if (IPCSocket != 0) return file_already_exists;
  sprintf(ipcfn,"%s%s.ipc",IPC_PATH,pipeName);
  if ((IPCSocket=CMDlisten(ipcfn)) == 0) return file_error;
  return 0;
}
#ifdef MULTITHREADED
MXHANDLE hCyclerStartMutex = 0;
MXHANDLE hCyclerStopMutex = 0;
DWORD dwTmrThreadId;
HANDLE hTmrThread = 0;
void afun(void)
{
  static int cnt = 0;
  cnt++;
  return;
}
THREAD_TYPE tmrTaskThread(THREAD_PARAM lpvThreadParam)
{
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0) 
  {
    feclog("tmrTaskThread: start cycler mutex unavailable !");
    return semaphore_error;
  }
  feclog("timer thread starting");
  while (!cycleTerminated)
  {
    // if nobody's at home, the pause a bit to spare the CPU 
    if (!ServerInitialized && nConnectionTableEntries == 0) millisleep(gSystemDelay);
    // run thru SystemCycle()
    srvTmrTask();
    // SystemCycle() should block on select with timeout of gSystemDelay
    //if (cycle_busy) millisleep(gSystemDelay);
    // let windows do its thing ...    
    if (!cycleTerminated) 
    {

      DoEvents(50);
    }
  }
  cycleStarted = FALSE;
  feclog("timer thread ending");
  feclog("shutting down server ports");
  SystemCloseServers();
  ReleaseSystemMutex(hCyclerStartMutex);
  return 0;
}
HANDLE tmrCreateThread(void)
{
  HANDLE hThread = 0;
  int threadPriority = -1;
  char *ptr = getenv("TINE_CYCLE_THREAD_PRIORITY");

  cycleTerminated = FALSE;
  hThread = CreateThread(NULL,0,tmrTaskThread,0,0,&dwTmrThreadId);
  if (ptr != NULL) threadPriority = atoi(ptr);
  if (threadPriority > 0) 
  {
    feclog("cycle thread priority set to %d via environment",threadPriority);
    SetThreadPriority(hThread,threadPriority);
  }
  return hThread;
}
int isUsingWindowsCycleTimer = FALSE;
#else
int hCyclerStartMutex = 0;
int hCyclerStopMutex = 0;
int isUsingWindowsCycleTimer = TRUE;
#endif

__declspec(dllexport) LRESULT CALLBACK TineCyleWndProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  switch (iMessage)
  {
    case WM_TIMER:       
      srvTmrTask(); 
      if (!gSelectOnSysPoll ||   // select fell through
          ReturnToClientCycle || // client loop needs attention
          ReturnToServerCycle)   // server loop needs attention
        PostMessage(hTineCycleWnd,WM_TIMER,0,0);
      break;
    case WM_DESTROY:     
      KillTimer(hWnd,1);
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}

int CreateTineCycleWindow(void)
{
  static WNDCLASS wndclass;
  static char  lpszWN[256];

  strncpy(lpszWN,gTineCycleWindowName,256); lpszWN[256] = 0;
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS;  // essential style!!
  wndclass.lpfnWndProc   = (WNDPROC) TineCyleWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hParent;
  wndclass.hIcon         = LoadIcon (hNULL, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (hNULL, IDC_ARROW);
  wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = hNULL;
  wndclass.lpszClassName = lpszWN;

  if (!RegisterClass(&wndclass)) return -1;

  if (!(hTineCycleWnd = CreateWindow (
			lpszWN, lpszWN,               // lpClassName, lpWindowName
			WS_BORDER|WS_CAPTION,         // dwStyle
			CW_USEDEFAULT, CW_USEDEFAULT, // X, Y
			CW_USEDEFAULT, CW_USEDEFAULT, // nWidth, nHeight
			hNULL, hNULL,                 // hWndParent, hMenu
			hParent,                      // hInstance
			(LPSTR)""                     // createstruct
      ) ) ) return -1;

  feclog("created timer Message Loop");
  return 0;
}

VOID CALLBACK srvTimerProc(
  HWND hwnd,         // handle to window
  UINT uMsg,         // WM_TIMER message
  UINT_PTR idEvent,  // timer identifier
  DWORD dwTime       // current system time
)
{
  srvTmrTask();
}

TINE_EXPORT void SystemSetCycleTimer(void)
{
  int cc = 0;
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemSetCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if (cycleStarted) 
  {
    feclog("attempt to start cycler while it is still in use");
    ccerr(-1);
  }
  cycleStarted = TRUE;
  feclog("starting cycler");
  if (isUsingWindowsCycleTimer)
  {
    CreateTineCycleWindow();
    idCycleTimer = SetTimer(hTineCycleWnd,1,gSystemDelay,NULL);
    //idCycleTimer = SetTimer(NULL,0,gSystemDelay,(TIMERPROC)srvTimerProc);
    if (idCycleTimer == 0) 
    {
      feclog("failed to create cycle timer : %d",GetLastError());
    }
  }
# ifdef MULTITHREADED
  else
  {
    feclog("creating cycler thread");
    UseMultiThreadedCycler(TRUE);
    hTmrThread = tmrCreateThread();
    millisleep(gSystemDelay);
  }
# endif
err:
  ReleaseSystemMutex(hCyclerStopMutex);
}
TINE_EXPORT void SystemKillCycleTimer(void)
{
  int cc = 0;
  if (WaitForSystemMutex(hCyclerStopMutex,-1) != 0) 
  {
    feclog("SystemKillCycleTimer: stop cycler mutex unavailable !");
    return;
  }
  if (!cycleStarted || cycleTerminated) 
  {
    feclog("attempt to kill timer refused : cycle %s",cycleStarted ? "already terminated" : "not started");
    ReleaseSystemMutex(hCyclerStopMutex);
    return;
  }
  feclog("terminating cycle");
  cycleTerminated = TRUE;
# ifdef MULTITHREADED
  if (!isUsingWindowsCycleTimer)
  {
    millisleep(gSystemDelay*2);
  }
# endif
  ReleaseSystemMutex(hCyclerStopMutex);
  if (WaitForSystemMutex(hCyclerStartMutex,-1) != 0) 
  {
    //feclog("SystemKillCycleTimer: start cycler mutex unavailable !");
    return;
  }
  ReleaseSystemMutex(hCyclerStartMutex);
}
TINE_EXPORT int SetSystemDelay(int msecs)
{
  if (msecs >= 10) gSystemDelay = msecs;
  if (gSystemDelay < MaxPollingRate)
  {
    MaxPollingRate = MinPollingRate = gSystemDelay;
  }
  return gSystemDelay;
}
int hasMessages(void)
{
  DWORD dw = GetQueueStatus(QS_SENDMESSAGE|QS_POSTMESSAGE);
  return dw == 0 ? FALSE : TRUE;
}
void DoEvents(int maxevents)
{
  MSG msg;
  int i = 0;
  while (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
    if (maxevents > 0 && i++ > maxevents) break;
  }
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
# ifdef MULTITHREADED
  char *ptr = getenv("TINE_USE_TIMER");
# endif
  struct _timeb tb;
# ifdef MULTITHREADED
  if (ptr != NULL)
  { // if not set use the default (currently set to TRUE)
    if (!stricmp(ptr,"YES") || !stricmp(ptr,"TRUE")) isUsingWindowsCycleTimer = TRUE;
    if (!stricmp(ptr,"NO") || !stricmp(ptr,"FALSE")) isUsingWindowsCycleTimer = FALSE;
  }
# endif
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      _ftime(&tb);
      sprintf(gTineCycleWindowName,"TineCycle @ %d:%d",tb.time,tb.millitm);
#     ifdef MULTITHREADED
      InitMutex(&hCyclerStartMutex);
      InitMutex(&hCyclerStopMutex);
#     endif
      hParent = hInstDll;
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      if (hTineCycleWnd != NULL)
      {
        DestroyWindow (hTineCycleWnd);
        UnregisterClass(gTineCycleWindowName,hParent);
      }
      DestroyShmWindow();
      break;
  }
  return (TRUE);
}
